from .material import Material
from .utils import check_input, is_positive

class Layer(Material):
    """ Represents a layer in target

    Parameters
    ----------
    elements : :obj:`dict`
        dictionary of elements (:class:`srim.core.elements.Element`, :obj:`str`, or :obj:`int`) with properties
         - ``stoich``  (float, int, required): Stoichiometry of element (fraction)
         - ``E_d``     (float, int, optional): Displacement energy [eV] default 25.0 eV
         - ``lattice`` (float, int, optional): Lattice binding energies [eV] default 0.0 eV
         - ``surface`` (float, int, optional): Surface binding energies [eV] default 3.0 eV
    density : :obj:`float`
       density [g/cm^3] of material
    width : :obj:`float`
       width [Angstroms] of layer
    phase : :obj:`int`
       phase of material (solid = 0, gas = 1). Default solid (0).
    name : :obj:`str:, optional
       name of the Layer (defaults to chemical_formula)

    Examples
    --------
    Construct a layer of SiC with experimental values.

    >>> Layer({
        'Si': {
           'stoich': 0.5,
           'E_d': 35.0, # Displacement Energy [eV]
           'lattice': 0.0,
           'surface': 3.0
        },
        'C': {
           'stoich': 0.5,
           'E_d': 20.0, # Displacement Energy [eV]
           'lattice': 0.0,
           'surface': 3.0
    }, density=3.21, width=10000.0)
    """
    def __init__(self, elements, density, width, phase=0, name=None):
        """Creation of Layer from elements, density, width, phase, and
name"""
        self.width = width
        self.name = name
        super(Layer, self).__init__(elements, density, phase)

    @classmethod
    def from_formula(cls, chemical_formula, density, width, phase=0, name=None):
        """ Creation Layer from chemical formula string, density, width, phase, and name

        Parameters
        ----------
        chemical_formula : str
            see :meth:`srim.core.material.Material.from_formula` for
            allowed formulas. Quite flexible.
        density : :obj:`float`
            density [g/cm^3] of material
        width : :obj:`float`
            width [Angstroms] of layer
        phase : :obj:`int`
            phase of material (solid = 0, gas = 1). Default solid (0).
        name : :obj:`str:, optional
            name of the Layer (defaults to chemical_formula)

        Notes
        -----
            This method is not used as much since you do not have an
            easy way to set the displacement energy.
        """
        elements = cls._formula_to_elements(chemical_formula)
        return Layer(elements, density, width, phase, name)

    @property
    def width(self):
        """Layer's width"""
        return self._width

    @width.setter
    def width(self, value):
        self._width = check_input(float, is_positive, value)

    @property
    def name(self):
        """Layer's Name"""
        if self._name:
            return self._name
        return self.chemical_formula

    @name.setter
    def name(self, value):
        self._name = str(value)

    def __repr__(self):
        return "<Layer material:{} width:{}>".format(self.chemical_formula, self.width)



""" Module for automating srim calculations

"""
import os
import subprocess
import shutil
import distutils.spawn

from .core.utils import (
    check_input,
    is_zero, is_zero_or_one, is_zero_to_two, is_zero_to_five,
    is_one_to_seven, is_one_to_eight,
    is_srim_degrees,
    is_positive,
    is_quoteless
)

from .output import Results
from .input import AutoTRIM, TRIMInput, SRInput
from .config import DEFAULT_SRIM_DIRECTORY


class TRIMSettings(object):
    """ TRIM Settings

    This object can construct all options available when running a TRIM calculation.

    Parameters
    ----------
    description : :obj:`str`, optional
       A name to give calculation. Has no effect on the actual
       calculation.
    reminders : :obj:`str`, optional
       TODO: could not find description. default 0
    autosave : :obj:`int`, optional
       save calculations after every `autosave` steps. default 0 will
       not autosave except at end
    plot_mode : :obj:`int`, optional
       Default 5.
       (0) ion distribution with recoils projected on y-plane
       (1) ion distribution with recoils projected on z-plane
       (2) ion distribution without recoils projected on y-plane
       (3) transverse plot of ions + recoil cascades, yz-plane
       (4) all four (0-3) on one screen
       (5) no graphics (default and at least 5X faster than others)
    plot_xmin : :obj:`float`, optional
       minimum x depth to plot only really matters if ``plot_mode``
       between 0-4. Default 0.0.
    plot_xmax : :obj:`float`, optional
       maximum x depth to plot only really matters if ``plot_mode``
       between 0-4. Default 0.0.
    ranges : :obj:`bool`, optional
       whether include ``RANGES.txt``, ``RANGE_3D.txt`` to output
       files. Default (0) False
    backscattered : :obj:`bool`, optional
       whether include ``BACKSCAT.txt`` to output files. Default (0)
       False
    transmit : :obj:`bool`, optional
       whether include ``TRANSMIT.txt`` to output files. Default (0)
       False
    sputtered : :obj:`bool`, optional
       whether include ``SPUTTER.txt`` to output files. Default (0)
       False
    collisions : :obj:`bool`, optional
       whether include ``COLLISON.txt`` to output files. Yes they did
       mispell collisions. Default (0) False
    exyz : int
       increment in eV to use for ``EXYZ.txt`` file. Default (0)
    angle_ions : :obj:`float`, optional
       angle of incidence of the ion with respect to the target
       surface. Default (0) perpendicular to the target surface along
       x-axis. Values 0 - 89.9.
    bragg_correction : :obj:`float`, optional
       bragg correction to stopping. Default (0) no correction
    random_seed : :obj:`int`, optional
       a random seed to start calculation with. Default (0) results
       should be reproducible (not always what you want)
    version : :obj:`int`, optional
       SRIM-2008 or SRIM-2008 so not really much choice. Default (0)

    Notes
    -----
        This class should never explicitely created. Instead set as
        kwargs in :class:`srim.srim.TRIM`
    """
    def __init__(self, **kwargs):
        """Initialize settings for a TRIM running"""
        self._settings = {
            'description': check_input(str, is_quoteless, kwargs.get('description', 'pysrim run')),
            'reminders': check_input(int, is_zero_or_one, kwargs.get('reminders', 0)),
            'autosave': check_input(int, is_zero_or_one, kwargs.get('autosave', 0)),
            'plot_mode': check_input(int, is_zero_to_five, kwargs.get('plot_mode', 5)),
            'plot_xmin': check_input(float, is_positive, kwargs.get('plot_xmin', 0.0)),
            'plot_xmax': check_input(float, is_positive, kwargs.get('plot_xmax', 0.0)),
            'ranges': check_input(int, is_zero_or_one, kwargs.get('ranges', 0)),
            'backscattered': check_input(int, is_zero_or_one, kwargs.get('backscattered', 0)),
            'transmit': check_input(int, is_zero_or_one, kwargs.get('transmit', 0)),
            'sputtered': check_input(int, is_zero_or_one, kwargs.get('ranges', 0)),
            'collisions': check_input(int, is_zero_to_two, kwargs.get('collisions', 0)),
            'exyz': check_input(int, is_positive, kwargs.get('exyz', 0)),
            'angle_ions': check_input(float, is_srim_degrees, kwargs.get('angle_ions', 0.0)),
            'bragg_correction': float(kwargs.get('bragg_correction', 1.0)), # TODO: Not sure what correct values are
            'random_seed': check_input(int, is_positive, kwargs.get('random_seed', 0)),
            'version': check_input(int, is_zero_or_one, kwargs.get('version', 0)),
        }

        if self.plot_xmin > self.plot_xmax:
            raise ValueError('xmin must be <= xmax')

    def __getattr__(self, attr):
        return self._settings[attr]


class TRIM(object):
    """ Automate TRIM Calculations

    Parameters
    ----------
    target : :class:`srim.core.target.Target`
        constructed target for TRIM calculation
    ion : :class:`srim.core.ion.Ion`
        constructed ion for TRIM calculation
    calculation : :obj:`int`, optional
        Default 1 full cascade calculation
        (1) Ion Distribution and Quick Calculation of Damage (quick KP)
        (2) Detailed Calculation with full Damage Cascades (full cascades)
        (3) Monolayer Collision Steps / Surface Sputtering
        (4) Ions with specific energy/angle/depth (quick KP damage) using TRIM.DAT
        (5) Ions with specific energy/angle/depth (full cascades) using TRIM.DAT
        (6) Recoil cascades from neutrons, etc. (full cascades) using TRIM.DAT
        (7) Recoil cascades and monolayer steps (full cascades) using TRIM.DAT
        (8) Recoil cascades from neutrons, etc. (quick KP damage) using TRIM.DAT
    number_ions : :obj:`int`, optional
        number of ions that you want to simulate. Default 1000. A lot
        better than the 99999 default in TRIM...
    kwargs :
        See :class:`srim.srim.TRIMSettings` for available TRIM
        options. There are many and none are required defaults are
        appropriate for most cases.

    Notes
    -----
        If you are doing a simulation with over 1,000 ions it is
        recomended to split the calculaion into several smaller
        calculations. TRIM has been known to unexpectedly crash mainly
        due to memory usage.
    """
    def __init__(self, target, ion, calculation=1, number_ions=1000, **kwargs):
        """ Initialize TRIM calcualtion"""
        self.settings = TRIMSettings(**kwargs)
        self.calculation = check_input(int, is_one_to_seven, calculation)
        self.number_ions = check_input(int, is_positive, number_ions)
        self.target = target
        self.ion = ion

    def _write_input_files(self):
        """ Write necissary TRIM input files for calculation """
        AutoTRIM().write()
        TRIMInput(self).write()

    @staticmethod
    def copy_output_files(src_directory, dest_directory, check_srim_output=True):
        """Copies known TRIM files in directory to destination directory

        Parameters
        ----------
        src_directory : :obj:`str`
            source directory to look for TRIM output files
        dest_directory : :obj:`str`
            destination directory to copy TRIM output files to
        check_srim_output : :obj:`bool`, optional
            ensure that all files exist
        """
        known_files = {
            'TRIM.IN', 'PHONON.txt', 'E2RECOIL.txt', 'IONIZ.txt',
            'LATERAL.txt', 'NOVAC.txt', 'RANGE.txt', 'VACANCY.txt',
            'COLLISON.txt', 'BACKSCAT.txt', 'SPUTTER.txt',
            'RANGE_3D.txt', 'TRANSMIT.txt', 'TRIMOUT.txt'
        }

        if not os.path.isdir(src_directory):
            raise ValueError('src_directory must be directory')

        if not os.path.isdir(dest_directory):
            raise ValueError('dest_directory must be directory')

        for known_file in known_files:
            if os.path.isfile(os.path.join(
                    src_directory, known_file)):
                shutil.copy(os.path.join(
                    src_directory, known_file), dest_directory)
            elif os.path.isfile(os.path.join(src_directory, 'SRIM Outputs', known_file)) and check_srim_output:
                shutil.copy(os.path.join(
                    src_directory, 'SRIM Outputs', known_file), dest_directory)

    def run(self, srim_directory=DEFAULT_SRIM_DIRECTORY):
        """Run configured srim calculation

        This method:
         - writes the input file to ``<srim_directory/SR Module/TRIM.IN``
         - launches ``<srim_directory>/SR Module/SRModule.exe``. Uses ``wine`` if available (needed for linux and osx)

        Parameters
        ----------
        srim_directory : :obj:`str`, optional
            path to srim directory. ``SRIM.exe`` should be located in
            this directory. Default ``/tmp/srim/`` will absolutely
            need to change for windows.
        """
        current_directory = os.getcwd()
        try:
            os.chdir(srim_directory)
            self._write_input_files()
            # Make sure compatible with Windows, OSX, and Linux
            # If 'wine' command exists use it to launch TRIM
            if distutils.spawn.find_executable("wine"):
                subprocess.check_call(['wine', str(os.path.join('.', 'TRIM.exe'))])
            else:
                subprocess.check_call([str(os.path.join('.', 'TRIM.exe'))])
            os.chdir(current_directory)
            return Results(srim_directory)
        finally:
            os.chdir(current_directory)


class SRSettings(object):
    """ SR Settings

    Parameters
    ----------
    energy_min : :obj:`float`, optional
       lowest energy in [eV] to calculation range
    output_type : :obj:`int`, optional
       specify units for output table
       (1) eV/Angstrom
       (2) keV/micron
       (3) MeV/mm
       (4) keV / (ug/cm2)
       (5) MeV / (mg/cm2)
       (6) keV / (mg/cm2)
       (7) eV / (1E15 atoms/cm2)
       (8) L.S.S reduced units
    output_filename : :obj:`str`, optional
       filename to give for SR output from calcualtion
    correction : :obj:`float`, optional
       Bragg rule correction. Usually no correction needed for heavy
       elements. Default 1.0 implies 100% of value (no change). 1.1
       will increase by 10%.

    Notes
    -----
        This class should never explicitely created. Instead set as
        kwargs in :class:`srim.srim.SR`
    """
    def __init__(self, **args):
        self._settings = {
            'energy_min': check_input(float, is_positive, args.get('energy_min', 1.0E3)),
            'output_type': check_input(int, is_one_to_eight, args.get('output_type', 1)),
            'output_filename': args.get('output_filename', 'SR_OUTPUT.txt'),
            'correction': check_input(float, is_positive, args.get('correction', 1.0))
        }

    def __getattr__(self, attr):
        return self._settings[attr]


class SR(object):
    """ Automate SR Calculations

    Parameters
    ----------
    leyer : :class:`srim.core.layer.Layer`
        constructed layer for SR calculation
    ion : :class:`srim.core.ion.Ion`
        constructed ion for SR calculation
    kwargs :
        See :class:`srim.srim.SRSettings` for available SR
        options. There are a few and none are required. Defaults are
        appropriate for most cases.
    """
    def __init__(self, layer, ion, **kwargs):
        self.settings = SRSettings(**kwargs)
        self.layer = layer
        self.ion = ion

    def _write_input_file(self):
        """ Write necissary SR input file for calculation """
        SRInput(self).write()

    def run(self, srim_directory=DEFAULT_SRIM_DIRECTORY):
        """Run configured srim calculation

        This method:
         - writes the input file to ``<srim_directory/SR Module/TRIM.IN``
         - launches ``<srim_directory>/SR Module/SRModule.exe``. Uses ``wine`` if available (needed for linux and osx)

        Parameters
        ----------
        srim_directory : :obj:`str`, optional
            path to srim directory. ``SRIM.exe`` should be located in
            this directory. Default ``/tmp/srim`` will absolutely need
            to be changed for windows.
        """
        current_directory = os.getcwd()
        try:
            os.chdir(os.path.join(srim_directory, 'SR Module'))
            self._write_input_file()
            # Make sure compatible with Windows, OSX, and Linux
            # If 'wine' command exists use it to launch TRIM
            if distutils.spawn.find_executable("wine"):
                subprocess.check_call(['wine', str(os.path.join('.', 'SRModule.exe'))])
            else:
                subprocess.check_call([str(os.path.join('.', 'SRModule.exe'))])
        finally:
            os.chdir(current_directory)



from math import sqrt

from . import units
from .element import Element

class Ion(Element):
    """ Representation of ion traveling through medium

    Similar to :class:`srim.core.element.Element` but associates an
    energy with the element.

    Parameters
    ----------
    identifier : :obj:`str`, :obj:`int`
        Symbol, Name, or Atomic Number of ion
    energy : :obj:`float`
        Energy [eV] of ion
    mass : :obj:`float`, optional
        Mass [amu] of element. Default is most common isotope atomic
        weight

    Examples
    --------
    Constructing a Helium Ion.

    >>> Ion('He', 1e6)
    "<Ion element:He mass:4.00 energy:1.00e6 eV>"

    >>> Ion('He', energy=1e6, mass=4.2)
    "<Ion element:He mass:4.20 energy:1.00e6 eV>"
    """
    def __init__(self, identifier, energy, mass=None):
        """Initialize Ion"""
        if energy <= 0.0:
            raise ValueError('energy {} cannot be 0.0 or less'.format(energy))

        self._energy = energy
        super(Ion, self).__init__(identifier, mass)

    def __repr__(self):
        return "<Ion element:{} mass:{:2.2f} energy:{%1.2E} eV>".format(
            self.name, self.mass, self.energy)

    @property
    def energy(self):
        """Ion's energy [eV]"""
        return self._energy

    @property
    def velocity(self):
        """Ion's velocity [m/s]"""
        return sqrt(2 * (self.energy * units.eV) / (self.mass * units.amu))



""" Read output files of SRIM simulation

TODO: Read header information
"""
import os
import re
from io import BytesIO

import numpy as np

from .core.ion import Ion

# Valid double_regex 4, 4.0, 4.0e100
double_regex = r'[-+]?\d+\.?\d*(?:[eE][-+]?\d+)?'
symbol_regex = r'[A-Z][a-z]?'
int_regex = '[+-]?\d+'


class SRIMOutputParseError(Exception):
    """SRIM error reading output file"""
    pass


class SRIM_Output(object):
    def _read_name(self, output):
        raise NotImplementedError()

    def _read_ion(self, output):
        ion_regex = 'Ion\s+=\s+({})\s+Energy\s+=\s+({})\s+keV'.format(
            symbol_regex, double_regex)
        match = re.search(ion_regex.encode('utf-8'), output)
        if match:
            symbol = str(match.group(1).decode('utf-8'))
            energy = float(match.group(2)) #keV
            return Ion(symbol, 1000.0 * energy)
        raise SRIMOutputParseError("unable to extract ion from file")

    def _read_target(self, output):
        match_target = re.search(b'(?<=====\r\n)Layer\s+\d+\s+:.*?(?=====)', output, re.DOTALL)
        if match_target:
            print(match_target.group(0))
            layer_regex = (
                'Layer\s+(?P<i>\d+)\s+:\s+(.+)\r\n'
                'Layer Width\s+=\s+({0})\s+A\s+;\r\n'
                '\s+Layer #\s+(?P=i)- Density = ({0}) atoms/cm3 = ({0}) g/cm3\r\n'
                '((?:\s+Layer #\s+(?P=i)-\s+{1}\s+=\s+{0}\s+Atomic Percent = {0}\s+Mass Percent\r\n)+)'
            ).format(double_regex, symbol_regex)
            layers = re.findall(layer_regex.encode('utf-8'), match_target.group(0))
            if layers:
                element_regex = (
                    '\s+Layer #\s+(\d+)-\s+({1})\s+=\s+({0})\s+Atomic Percent = ({0})\s+Mass Percent\r\n'
                ).format(double_regex, symbol_regex)
                element_regex = element_regex.encode()

                layers_elements = []
                for layer in layers:
                    # We know that elements will match
                    layers_elements.append(re.findall(element_regex, layer[5]))

                raise NotImpementedError()

                import pytest
                pytest.set_trace()

        raise SRIMOutputParseError("unable to extract total target from file")

    def _read_num_ions(self, output):
        match = re.search(b'Total Ions calculated\s+=(\d+.\d+)', output)
        if match:
            # Cast string -> float -> round down to nearest int
            return int(float(match.group(1)))
        raise SRIMOutputParseError("unable to extract total ions from file")

    def _read_table(self, output):
        match = re.search((
            b'=+(.*)'
            b'-+(?:\s+-+)+'
        ), output, re.DOTALL)
        # Read Data from table

        if match:
            # Headers TODO: name the columns in table
            header = None

            # Data
            data = np.genfromtxt(BytesIO(output[match.end():]), max_rows=100)
            return data
        raise SRIMOutputParseError("unable to extract table from file")


class Results(object):
    """ Gathers all results from folder

    Parameters
    ----------
    directory : :obj:`str`
        directory to look for TRIM calculations

    Notes
    -----
    Files that are looked for:
      - ``IONIZ.txt`` handled by :class:`srim.output.Ioniz`
      - ``VACANCY.txt`` handled by :class:`srim.output.Vacancy`
      - ``NOVAC.txt`` handled by :class:`srim.output.NoVacancy`
      - ``E2RECOIL.txt`` handled by :class:`srim.output.EnergyToRecoils`
      - ``PHONON.txt`` handled by :class:`srim.output.Phonons`
      - ``RANGE.txt`` handled by :class:`srim.output.Range`
    """
    def __init__(self, directory):
        """ Retrives all the calculation files in a given directory"""
        self.ioniz = Ioniz(directory)
        self.vacancy = Vacancy(directory)

        try:
            self.novac = NoVacancy(directory)
        except ValueError:
            self.novac = None

        self.etorecoils = EnergyToRecoils(directory)
        self.phonons = Phonons(directory)
        self.range = Range(directory)


class Ioniz(SRIM_Output):
    """``IONIZ.txt`` Ionization by ions and depth. Includes header information about calculation

    Parameters
    ----------
    directory : :obj:`str`
         directory of calculation
    filename : :obj:`str`, optional
         filename for Ioniz. Default ``IONIZ.txt``
    """
    def __init__(self, directory, filename='IONIZ.txt'):
        with open(os.path.join(directory, filename), 'rb') as f:
            output = f.read()
            ion = self._read_ion(output)
            num_ions = self._read_num_ions(output)
            data = self._read_table(output)

        self._ion = ion
        self._num_ions = num_ions
        self._depth = data[:, 0]
        self._ions = data[:, 1]
        self._recoils = data[:, 2]

    @property
    def ion(self):
        """ Ion used in SRIM calculation

        **mass** could be wrong
        """
        return self._ion

    @property
    def num_ions(self):
        """ Number of Ions in SRIM simulation """
        return self._num_ions

    @property
    def depth(self):
        """ Depth [Ang] of bins in SRIM Calculation """
        return self._depth

    @property
    def ions(self):
        """Ionization energy [eV/(Angstrom Ion)] lost to electronic stopping
        in incident ions"""
        return self._ions

    @property
    def recoils(self):
        """Ionization energy [eV/(Angstrom Ion)] lost to electronic stopping
        in recoil ions"""
        return self._recoils


class Vacancy(SRIM_Output):
    """``VACANCY.txt`` Table of the final distribution of vacancies vs depth

    Parameters
    ----------
    directory : :obj:`str`
         directory of calculation
    filename : :obj:`str`, optional
         filename for Vacancy. Default ``VACANCY.txt``
    """
    def __init__(self, directory, filename='VACANCY.txt'):
        with open(os.path.join(directory, filename), 'rb') as f:
            output = f.read()
            ion = self._read_ion(output)
            num_ions = self._read_num_ions(output)
            data = self._read_table(output)

        self._ion = ion
        self._num_ions = num_ions
        self._depth = data[:, 0]
        self._ion_knock_ons = data[:, 1]
        self._vacancies = data[:, 2:]

    @property
    def ion(self):
        """ Ion used in SRIM calculation

        **mass** could be wrong
        """
        return self._ion

    @property
    def num_ions(self):
        """Number of Ions in SRIM simulation"""
        return self._num_ions

    @property
    def depth(self):
        """Depth [Ang] of bins in SRIM Calculation"""
        return self._depth

    @property
    def knock_ons(self):
        """Vacancies produced [Vacancies/(Angstrom-Ion) by ion]"""
        return self._ion_knock_ons

    @property
    def vacancies(self):
        """Vacancies [Vacancies/(Angstrom-Ion)] produced of element in layer"""
        return self._vacancies


class NoVacancy(SRIM_Output):
    """ ``NOVAC.txt`` Table of Replacement Collisions

    Parameters
    ----------
    directory : :obj:`str`
         directory of calculation
    filename : :obj:`str`, optional
         filename for NoVacancy. Default ``NOVAC.txt``
    """
    def __init__(self, directory, filename='NOVAC.txt'):
        with open(os.path.join(directory, filename), 'rb') as f:
            output = f.read()

            # Check if it is KP calculation
            if re.search(b'Recoil/Damage Calculations made with Kinchin-Pease Estimates',
                         output):
                raise ValueError('NOVAC has no data for KP calculations')

            ion = self._read_ion(output)
            num_ions = self._read_num_ions(output)
            data = self._read_table(output)

        self._ion = ion
        self._num_ions = num_ions
        self._depth = data[:, 0]
        self._number = data[:, 1]

    @property
    def ion(self):
        """ Ion used in SRIM calculation

        **mass** could be wrong
        """
        return self._ion

    @property
    def num_ions(self):
        """Number of Ions in SRIM simulation"""
        return self._num_ions

    @property
    def depth(self):
        """Depth [Ang] of bins in SRIM Calculation"""
        return self._depth

    @property
    def number(self):
        """Replacement Collisions [Number/(Angstrom-Ion)]"""
        return self._number


class EnergyToRecoils(SRIM_Output):
    """``E2RECOIL.txt`` Energy transfered to atoms through binary collision

    Parameters
    ----------
    directory : :obj:`str`
         directory of calculation
    filename : :obj:`str`, optional
         filename for EnergyToRecoils. Default ``E2RECOIL.txt``
    """
    def __init__(self, directory, filename='E2RECOIL.txt'):
        with open(os.path.join(directory, filename), 'rb') as f:
            output = f.read()
            ion = self._read_ion(output)
            num_ions = self._read_num_ions(output)
            data = self._read_table(output)

        self._ion = ion
        self._num_ions = num_ions
        self._depth = data[:, 0]
        self._ions = data[:, 1]
        self._recoils = data[:, 2:]

    @property
    def ion(self):
        """Ion used in SRIM calculation

        **mass** could be wrong
        """
        return self._ion

    @property
    def num_ions(self):
        """Number of Ions in SRIM simulation"""
        return self._num_ions

    @property
    def depth(self):
        """Depth [Ang] of bins in SRIM Calculation"""
        return self._depth

    @property
    def ions(self):
        """Energy [eV/(Angstrom-Ion)] transfered to material through ion collisions"""
        return self._ions

    @property
    def absorbed(self):
        """Energy [eV/(Angstrom-Ion)] absorbed from collisions with Atom

        TODO: fix terminology
        """
        return self._recoils


class Phonons(SRIM_Output):
    """``PHONON.txt``  Distribution of Phonons

    Parameters
    ----------
    directory : :obj:`str`
         directory of calculation
    filename : :obj:`str`, optional
         filename for Phonons. Default ``PHONON.txt``
    """
    def __init__(self, directory, filename='PHONON.txt'):
        with open(os.path.join(directory, filename), 'rb') as f:
            output = f.read()
            ion = self._read_ion(output)
            num_ions = self._read_num_ions(output)
            data = self._read_table(output)

        self._ion = ion
        self._num_ions = num_ions
        self._depth = data[:, 0]
        self._ions = data[:, 1]
        self._recoils = data[:, 2]

    @property
    def ion(self):
        """Ion used in SRIM calculation

        **mass** could be wrong
        """
        return self._ion

    @property
    def num_ions(self):
        """Number of Ions in SRIM simulation"""
        return self._num_ions

    @property
    def depth(self):
        """Depth [Ang] of bins in SRIM Calculation"""
        return self._depth

    @property
    def ions(self):
        """Number of phonons [Phonons/(Angstrom Ion)] created from ions collisions"""
        return self._ions

    @property
    def recoils(self):
        """Number of phonons [Phonons/(Angstrom Ion)] created from recoils
        resulting from ion collisions"""
        return self._recoils


class Range(SRIM_Output):
    """``RANGE.txt`` Table of the final distribution of the ions, and any recoiling target atoms

    Parameters
    ----------
    directory : :obj:`str`
         directory of calculation
    filename : :obj:`str`, optional
         filename for Range. Default ``RANGE.txt``
    """
    def __init__(self, directory, filename='RANGE.txt'):
        with open(os.path.join(directory, filename), 'rb') as f:
            output = f.read()
            ion = self._read_ion(output)
            num_ions = self._read_num_ions(output)
            data = self._read_table(output)

        self._ion = ion
        self._num_ions = num_ions
        self._depth = data[:, 0]
        self._ions = data[:, 1]
        self._elements = data[:, 2:]

    @property
    def ion(self):
        """Ion used in SRIM calculation

        **mass** could be wrong
        """
        return self._ion

    @property
    def num_ions(self):
        """Number of Ions in SRIM simulation"""
        return self._num_ions

    @property
    def depth(self):
        """Depth [Ang] of bins in SRIM Calculation"""
        return self._depth

    @property
    def ions(self):
        """Ion final distribution [(Atoms/cm3)/(Atoms/cm2)]"""
        return self._ions

    @property
    def elements(self):
        """Per elements [(Atoms/cm3)/(Atoms/cm2)] distribution of each element"""
        return self._elements



class Backscat(object):
    """ The kinetics of all backscattered ions (energy, location and trajectory)

    TODO: one day to be implemented! submit pull request please!
    """
    pass


class Transmit(object):
    """ The kinetics of all transmitted ions (energy, location and trajectory)

    TODO: one day to be implemented! submit pull request please!
    """
    pass


class Sputter(object):
    """ The kinetics of all target atoms sputtered from the target.

    TODO: one day to be implemented! submit pull request please!
    """
    pass


class Collision:
    """Reads the SRIM Collisions file.

    This is the most important file in my opinion. It records every
    single collision and its energies. The file will get huge for
    simulations with many collisions. Since the file can be larger
    than the amount of RAM it will read the file in sections
    (buffers).

    Parameters
    ----------
    directory : :obj:`str`
         directory of calculation
    filename : :obj:`str`, optional
         filename for Collisions. Default ``COLLISON.txt``

    """
    def __init__(self, directory, filename='COLLISON.txt'):
        self.filename = os.path.join(directory, filename)

        with open(filename, encoding="latin-1") as f:
            self._read_header(f)

        self._ion_index = buffered_findall(filename, b"  Ion    Energy")

    def _read_header(self, f):
        """Read Header of COLLISON.txt

        Currently we do nothing with the header
        """

        # Collect the header of the file
        header = []

        for line in f:
            if line == " \n":
                break
            header.append(line)
        return header

    def _read_ion(self, ion_str):
        """There are 2 types of files with and without cascades

        format:
           1 - Kinchin-Pease Theory (No full cascades)
           2 - full cascades
        """
        # Notice that lines is an generator!
        # This makes it so we can walk through lines
        # in multiple for loops
        lines = (line for line in ion_str.split('\n'))

        # Skip Ion Header
        for line in lines:
            if re.match("^-+\r$", line):
                break

        collisions = []

        # Reads collisions for an ion
        for line in lines:
            if re.match("^=+\r$", line):
                break

            tokens = line.split(chr(179))[1:-1]

            # Check if a full_cascades simulation
            # Read Cascade information
            if re.match(r"\s+<== Start of New Cascade\s+", tokens[-1]):
                (target_disp,
                 target_vac,
                 target_replac,
                 target_inter,
                 cascade) = self._read_cascade(lines)
            else:
                target_disp = float(tokens[8])
                target_vac = 0
                target_replac = 0
                target_inter = 0
                cascade = None

            collisions.append({
                'ion_number': int(tokens[0]),
                'kinetic_energy': float(tokens[1]),
                'depth': float(tokens[2]),
                'lat_y_dist': float(tokens[3]),
                'lat_z_dist': float(tokens[4]),
                'stopping_energy': float(tokens[5]),
                'atom': re.search("([A-Z][a-z]?)", tokens[6]).group(1),
                'recoil_energy': float(tokens[7]),
                'target_disp': target_disp,
                'target_vac': target_vac,
                'target_replac': target_replac,
                'target_inter': target_inter,
                'cascade': cascade
            })

            # Handles weird case where no summary of cascade
            if target_disp is None:
                break;

        # Reads ion footer
        ion_number = re.search(int_regex, next(lines)).group(0)

        footer = ""
        for line in lines:
            if re.match("^=+\r$", line):
                break
            footer += line

        matches = re.findall(double_regex, footer)

        line = next(lines)

        return {
            'ion_number': int(ion_number),
            'displacements': float(matches[0]),
            'avg_displacements': float(matches[1]),
            'replacements': float(matches[2]),
            'avg_replacements': float(matches[3]),
            'vacancies': float(matches[4]),
            'avg_vacancies': float(matches[5]),
            'interstitials': float(matches[6]),
            'avg_interstitials': float(matches[7]),
            'sputtered_atoms': float(matches[8]),
            'avg_sputtered_atoms': float(matches[9]),
            'transmitted_atoms': float(matches[10]),
            'avg_transmitted_atoms': float(matches[11]),
            'collisions': collisions
        }

    def _read_cascade(self, lines):
        line = next(lines)

        assert re.match("^=+\r$", line)


        line = next(lines)
        assert re.match((
                "  Recoil Atom Energy\(eV\)   X \(A\)      Y \(A\)      Z \(A\)"
                "   Vac Repl Ion Numb \d+="
        ), line)

        cascade = []
        for line in lines:
            if re.match("^=+\r$", line):
                break
            tokens = line.split()[1:-1]

            print(tokens)
            cascade.append({
                'recoil': int(tokens[0]),
                'atom': int(tokens[1]),
                'recoil_energy': float(tokens[2]),
                'position': np.array([float(tokens[3]),
                                      float(tokens[4]),
                                      float(tokens[5])]),
                'vac': int(tokens[6]),
                'repl': int(tokens[7])
            })

        if line.count('=') > 100:
            return None, None, None, None, cascade

        line = next(lines)
        tokens = line.split(chr(179))[1:-1]

        if tokens:

            target_disp = float(tokens[2])
            target_vac = float(tokens[3])
            target_replac = float(tokens[4])
            target_inter = float(tokens[5])
        else:
            target_disp = None
            target_vac = None
            target_replac = None
            target_inter = None

        return target_disp, target_vac, target_replac, target_inter, cascade

    def __getitem__(self, i):
        start = self._ion_index[i]

        if i == len(self._ion_index):
            end = os.path.getsize(self.filename)
        else:
            end = self._ion_index[i+1]

        with open(self.filename, "rb") as f:
            f.seek(start)
            # We assume that ion_str will fit in RAM
            ion_str = f.read(end - start)
            return self._read_ion(ion_str.decode('latin-1'))

    def __len__(self):
        return len(self._ion_index) - 1


def buffered_findall(filename, string, start=0):
    """A method of reading a file in buffered pieces (needed for HUGE files)"""
    with open(filename, 'rb') as f:
        filesize = os.path.getsize(filename)
        BUFFERSIZE = 4096
        overlap = len(string) - 1
        buffer = None
        positions = []

        if start > 0:
            f.seek(start)

        while True:
            if (f.tell() >= overlap and f.tell() < filesize):
                f.seek(f.tell() - overlap)
            buffer = f.read(BUFFERSIZE)
            if buffer:
                buffer_positions = [m.start() for m in re.finditer(string, buffer)]

                for position in buffer_positions:
                    if position >= 0:
                        positions.append(f.tell() - len(buffer) + position)
            else:
                return positions



class Target(object):
    """ Target that Ion Impacts

    Parameters
    ----------
    layers : list
        list of :class:`srim.core.layer.Layer` to construct Target

    Examples
    --------
    Lets construct a SiC target. While only one layer is shown an
    arbitrary number of them can be used.

    >>> Target([Layer({
        'Si': {
           'stoich': 0.5,
           'E_d': 35.0, # Displacement Energy [eV]
           'lattice': 0.0,
           'surface': 3.0
        },
        'C': {
           'stoich': 0.5,
           'E_d': 20.0, # Displacement Energy [eV]
           'lattice': 0.0,
           'surface': 3.0
    }, density=3.21, width=10000.0)])
    """
    def __init__(self, layers):
        self.layers = layers

    @property
    def width(self):
        """total width of target (sum of layers)"""
        return sum(layer.width for layer in self.layers)



def plot_damage_energy(results, ax):
    """Plot damage energy (ions + recoils) per unit depth

    Parameters
    ----------
    results : :class:`srim.output.Results`
        results from srim calcualtion
    ax : matplotlib.Axes
        matplotlib axes to plot into
    """
    phon = results['phonons']
    dx = max(phon.depth) / 100.0 # to units of Angstroms
    energy_damage = (phon.ions + phon.recoils) * dx
    ax.plot(phon.depth, energy_damage / phon.num_ions, label='{}'.format(folder))
    return sum(energy_damage)


def plot_ionization(results, ax):
    """Plot ionization (ion vs recoils) per unit depth

    Parameters
    ----------
    results : :class:`srim.output.Results`
        results from srim calcualtion
    ax : matplotlib.Axes
        matplotlib axes to plot into
    """
    ioniz = results['ioniz']
    dx = max(ioniz.depth) / 100.0 # to units of Angstroms
    ax.plot(ioniz.depth, ioniz.ions, label='Ionization from Ions')
    ax.plot(ioniz.depth, ioniz.recoils, label='Ionization from Recoils')


def plot_vacancies(results, ax):
    """Plot vacancies (ion + recoils produced) per unit depth

    Parameters
    ----------
    results : :class:`srim.output.Results`
        results from srim calcualtion
    ax : matplotlib.Axes
        matplotlib axes to plot into
    """
    vac = results['vacancy']
    vacancy_depth = vac.knock_ons + np.sum(vac.vacancies, axis=1)
    ax.plot(vac.depth, vacancy_depth, label="Total vacancies at depth")
    return sum(vacancy_depth)
